// Collector.java
// Author: Stuart Clayman
// Email: sclayman@ee.ucl.ac.uk
// Date: July 2009
// Update: Oct 2010

package eu.reservoir.aggregator;

import eu.reservoir.monitoring.core.*;
import eu.reservoir.monitoring.core.plane.DataPlane;
import eu.reservoir.monitoring.distribution.*;
import eu.reservoir.monitoring.appl.BasicConsumer;
import java.io.Serializable;
import java.io.File;
import java.util.Scanner;
import java.util.Properties;
import com.timeindexing.index.*;
import com.timeindexing.time.MillisecondTimestamp;
import com.timeindexing.data.SerializableItem;
import com.timeindexing.cache.*;

/**
 * A Collector is a consumer of measurements.
 * It saves the measurements for later selection.
 */
public class Collector  implements Reporter {
    /*
     * The Consumer of the Measurements.
     */
    BasicConsumer consumer = null;

    /*
     * The Time Index that holds the consumed measurements.
     */
    IndexView index;
    File indexFullPath = null;
    TimeIndexFactory factory = null;
    Properties indexProperties = null;

    /*
     * The current extractor.
     * Start with identity function.
     */
    Extractor extractor = new Extractor() {
	    public Serializable extract(Measurement m) {
		return (Serializable)m;
	    }
	};


    /*
     * The data plane
     */
    DataPlane dataPlane;

    /**
     * The Collector listens to a DataPlane and stores the collected
     * measurements in a timeindex called indexName at path indexPath.
     * This gets measurements as a MeasurementReceiver
     */
    public Collector(DataPlane dataPlane, String indexName, String indexPath) {
	// Set up a Consumer
        this.dataPlane = dataPlane;
	consumer = new BasicConsumer();

        // set up data plane
        consumer.setDataPlane(dataPlane);


	consumer.setReporter(this);

        // Setup TimeIndex to store data
        factory = new TimeIndexFactory();
        indexProperties = new Properties();

        String realName = indexName+"-collected";
        indexFullPath = new File(indexPath, realName);
        indexProperties.setProperty("indexpath",  indexFullPath.getPath());
        indexProperties.setProperty("name", realName);

	try {
	    // Create TimeIndex to store data
	    index = factory.create(IndexType.EXTERNAL, indexProperties);

            index.setCachePolicy(new HollowAtDataVolumePolicy(1024*1024));
            index.setAutoCommit(true);
	} catch (TimeIndexException tie) {
	    tie.printStackTrace();
	    throw new RuntimeException("Cannot create TimeIndex ");
	}

	// the index is now open


    }

    /**
     * Receiver of a measurement.
     * They are stored for later use by the Selector and the Aggregator.
     */
    public void report(Measurement m) {
	// original send time
	long sendTimeMillis = m.getTimestamp().value();

	// save the measurement
	Serializable serializable = extractor.extract(m);

	try {
            if (!index.isClosed()) {
                // do not save with sender timestamp
                // as this can go wrong with multiple producers.
                // index.addItem(new SerializableItem(serializable), new MillisecondTimestamp(sendTimeMillis));
                index.addItem(new SerializableItem(serializable));
            }
        } catch (TimeIndexException tie) {
            tie.printStackTrace();
            throw new RuntimeException("Can't add measurement to time index: " + index.getIndexPathName());
        }
    }

    /**
     * Start the Collector.
     */
    public void start() {
	// we connect to the network
	consumer.connect();
	System.out.println("Collector connect to " + dataPlane);
	System.out.println("Collector index " + index.getIndexPathName());
    }

    /**
     * Stop the Collector.
     */
    public void stop() {
	// we disconnect from the network
	consumer.disconnect();
	System.out.println("Collector disconnect from " + dataPlane);

        try {
            index.close();
	} catch (TimeIndexException tie) {
	    tie.printStackTrace();
	}
    }


    /**
     * Get the current Extractor function.
     */
    public Extractor getExtractor() {
	return extractor;
    }

    /**
     * Set the  Extractor function to use;
     */
    public Collector setExtractor(Extractor e) {
	extractor = e;
	return this;
    }


    /**
     * Get a handle on the time index.
     */
    public IndexView getIndex() {
	return index;
    }


}
