// Forwarder.java
// Author: Stuart Clayman
// Email: sclayman@ee.ucl.ac.uk
// Date: July 2009

package eu.reservoir.aggregator;

import eu.reservoir.monitoring.core.*;
import eu.reservoir.monitoring.distribution.*;
//import eu.reservoir.monitoring.im.*;
import eu.reservoir.monitoring.core.plane.DataPlane;
import java.net.InetAddress;
import java.io.File;
import java.util.Scanner;
import java.util.Properties;
import java.util.List;
import com.timeindexing.index.*;
import com.timeindexing.time.MillisecondTimestamp;
import com.timeindexing.data.SerializableItem;
import com.timeindexing.cache.*;

/**
 * This forwards data to a multicast address.
 * Uses a DataSourceMulticaster to distribute the data.
 */
public class Forwarder {
    /*
     * The data plane is being forwared to.
     */
    DataPlane dataPlane;

    /*
     * The DataSource
     */
    DataSource ds;

    /* 
     * The Probe
     */
    Probe probe;

    /*
     * The Time Index that holds the forwared data.
     */
    IndexView forwardIndex;
    File forwardIndexPath = null;
    Properties forwardProperties = null;

    /*
     * The Time Index that holds the filtered data.
     */
    IndexView filteredIndex;
    File filterIndexPath = null;
    Properties filterProperties = null;

    TimeIndexFactory factory = null;

    /*
     * The Filter for the numbers.
     * The default one always returns true.
     */
    Filter filter = new Filter() {
	    public boolean filter(Forwarder f, AggregatorMeasurement m) {
		return true;
	    }
	};

    /*
     * The current value to forward.
     */
    AggregatorMeasurement value = null;

    /*
     * The previous value to forward.
     */
    AggregatorMeasurement oldValue = null;

    /**
     * Construct a Forwarder to a specified data plane.
     * Filtered and forwarded data are stored in separate timeindexes.
     * The forwarded data is passed as a measurement in a field called: fieldName.
     */
    public Forwarder(DataPlane dataPlane, String fieldName, String indexName, String indexPath, List<ProbeAttribute> probeAttributes) {
	// set up data source
	ds = new ForwarderDataSource();
        ds.setDataPlane(dataPlane);

	probe = new ForwarderProbe(this, probeAttributes);

	// add the probe to the data source
	ds.addProbe(probe);

	// set up DataPlane to DataSource
	this.dataPlane = dataPlane;

	// now set up the time indexes for logging
	if (indexName != null && indexPath != null) {
            // create a TimeIndexFactory`
            factory = new TimeIndexFactory();
            forwardProperties = new Properties();
            filterProperties = new Properties();

            String realName;

            // set up forwardIndex
            realName = indexName+"-forwarded";
            forwardIndexPath = new File(indexPath, realName);
            forwardProperties.setProperty("indexpath",  forwardIndexPath.getPath());
            forwardProperties.setProperty("name", realName);

            // set up filterIndex
            realName = indexName+"-filtered";
            filterIndexPath = new File(indexPath, realName);
            filterProperties.setProperty("indexpath",  filterIndexPath.getPath());
            filterProperties.setProperty("name", realName);

        }


    }

    /**
     * Construct a Forwarder to a specified data plane.
     * The forwarded data is passed as a measurement in a field called: fieldName.
     */
    public Forwarder(DataPlane dataPlane, String fieldName,  List<ProbeAttribute> probeAttributes) {
	this(dataPlane, fieldName, null, null, probeAttributes);
    }

    /**
     * Get the current value which has to be forwarded.
     */
    public AggregatorMeasurement getValue() {
	return value;
    }

    /**
     * Get the previous value which has to be forwarded.
     */
    public AggregatorMeasurement getOldValue() {
	    return oldValue;
    }

    /**
     * Set the current value to forward.
     * This informs the Probe of the value if the filter succeeds.
     * Sets the value just before return so Filters can use
     * getValue() to see previous value.
     */
    public Forwarder setValue(AggregatorMeasurement n) {
	if (filter.filter(this, n)) {
	    System.out.println("Forwarder: forward " + n);

	    // we only update the value if the
	    // filter returns true
	    // TODO: think??? maybe we should always update the value
	    oldValue = value;
	    value = n;

	    probe.inform(n);

	    // do loggging to time index
	    if (forwardIndex != null) {
		try {
		    forwardIndex.addItem(new SerializableItem(n));
		} catch (TimeIndexException tie) {
		    System.err.println("Can't add forward info to time index log " + forwardIndex.getName());
		}
	    }
	} else {
	    System.out.println("Forward: filtered " + n);

	    // do loggging to time index
	    if (filteredIndex != null) {
		try {
		    filteredIndex.addItem(new SerializableItem(n));
		} catch (TimeIndexException tie) {
		    System.err.println("Can't add forward info to time index log " + filteredIndex.getName());
		}
	    }

	}


	return this;
    }

    /**
     * Get the current Filter function.
     */
    public Filter getFilter() {
	return filter;
    }

    /**
     * Set the  Filter function to use;
     */
    public Forwarder setFilter(Filter f) {
	filter = f;
	return this;
    }

    /**
     * Start the Forwarder.
     */
    public void start() {
	// now set up the time indexes for logging
        try {
            forwardIndex = factory.create(IndexType.EXTERNAL, forwardProperties);
            forwardIndex.setCachePolicy(new HollowAtDataVolumePolicy(256*1024));
            forwardIndex.setAutoCommit(true);

            filteredIndex = factory.create(IndexType.EXTERNAL, filterProperties);
            filteredIndex.setCachePolicy(new HollowAtDataVolumePolicy(256*1024));
            filteredIndex.setAutoCommit(true);

        } catch (TimeIndexException tie) {
            tie.printStackTrace();
            throw new RuntimeException("Cannot create TimeIndex ");
        }

        // the index is now open


	// turn on the probe 
	ds.connect();
	System.out.println("Forwarder connect to " + dataPlane);
    }

    /**
     * Stop the Forwarder.
     */
    public void stop() {
        ds.removeProbe(probe);
	ds.disconnect();
	System.out.println("Forwarder disconnect from " + dataPlane);


        try {
            forwardIndex.close();
	} catch (TimeIndexException tie) {
	    tie.printStackTrace();
	}



        try {
            filteredIndex.close();
	} catch (TimeIndexException tie) {
	    tie.printStackTrace();
	}



    }
}

/**
 * A Data Source.
 */
class ForwarderDataSource extends AbstractDataSource {
  /*
   * Create a DataSource
   */
  public ForwarderDataSource() {
      // try and get the real current hostname
      String currentHost ="localhost";

      setName(currentHost + "ForwarderDataSource");
  }

}
