// TixSelect.java
// Author: Stuart Clayman
// Email: sclayman@ee.ucl.ac.uk
// Date: Sept 2009

package eu.reservoir.aggregator.demo;

import com.timeindexing.index.Index;
import com.timeindexing.index.IndexItem;
import com.timeindexing.index.IndexProperties;
import com.timeindexing.index.TimeIndexException;
import com.timeindexing.data.DataItem;
import com.timeindexing.time.Timestamp;
import com.timeindexing.time.MillisecondTimestamp;
import com.timeindexing.time.TimeCalculator;
import com.timeindexing.appl.ArgArrayConverter;
import com.timeindexing.appl.Selecter;
import com.timeindexing.appl.SelectionStreamer;
import com.timeindexing.plugin.OutputPlugin;
import com.timeindexing.plugin.WriterPlugin;
import com.timeindexing.plugin.DefaultOutputPlugin;
import com.timeindexing.plugin.DefaultWriter;
import java.io.OutputStream;
import java.io.PrintStream;
import java.io.IOException;
import eu.reservoir.monitoring.core.*;
import eu.reservoir.monitoring.distribution.*;
import eu.reservoir.monitoring.distribution.multicast.*;

/**
 * This class selects data from a Timeindex.
 */
public class TixSelect {
    Selecter select = null;

    String path;

    IndexProperties selectionProperties = null;

    public  TixSelect(String[] args) {
	/*
	 * Process args
	 */
	ArgArrayConverter argConverter = new ArgArrayConverter();
	
	selectionProperties = argConverter.convert(args);

	path = (String)selectionProperties.get("indexpath");
    }


    public void start() {
	// The WriterPlugin to use
	// Different plugins will show the output differently
	WriterPlugin writerPlugin = new TixStats();

	// create a Selecter to select stuff
	select = new Selecter(path, System.out, new DefaultOutputPlugin(writerPlugin));

	/*
	 * Go for it
	 */
	try {
	    // select from the Selecter
	    // passing in the properties
	    select.select(selectionProperties);
	} catch (TimeIndexException tie) {
	    System.err.println("TixSelect: error " + tie.getMessage());
	} catch (IOException ioe) {
	    System.err.println("TixSelect: error " + ioe.getMessage());
	} catch (Error e) {
	    try {
		select.close();
	    } catch (TimeIndexException tie) {
		System.err.println("Cannot close time index " + path);
	    }
	}
    
    }

    public void stop() {
    }

    public static void main(String [] args) {
	TixSelect selecter = new TixSelect(args);
	selecter.start();
    }

    /**
     * A WriterPlugin for doing time differences between
     * the index and the time of the Measurements.
     * The write() method gets called for each IndexItem.
     * The end() method is called at the end.
     * It calculates:
     * transmission time total
     * transmission volume
     * throughput
     * average transmission time
     * std dev of transmission time
     */
     class TixStats extends DefaultWriter {
	 // the total volume
	 long total = 0;

	 // the count
	 long count = 0;

	 // the sum
	 double sum = 0;

	 // the sum of squares
	 double sumSq = 0;

	 // The index
	 Index index = null;

	 public long write(IndexItem item, IndexProperties outputProperties) throws IOException {
	     // get the data item as something that can be deserialized
	     index = item.getIndex();
	     DataItem dataItem = item.getDataItem();
	     Object object = dataItem.getObject();

	     // we're outputting to System.out so getOutputStream() 
	     // should return a PrintStream
	     PrintStream out = (PrintStream)getOutputStream();

	     if (object instanceof ConsumerMeasurementWithMetaData) {
		 ConsumerMeasurementWithMetaData m = (ConsumerMeasurementWithMetaData)object;

		 // original send time
		 long sendTimeMillis = m.getTimestamp().value();
		 Timestamp sendTime = new MillisecondTimestamp(sendTimeMillis);

		 // arrival time in index
		 Timestamp arriveTime = item.getIndexTimestamp();

		 // and the difference is
		 Timestamp diff = TimeCalculator.subtractTimestamp(arriveTime, sendTime);
		 double transmissionTime = (((double)diff.getNanoSeconds()) / 1000000000);

		 MulticastTransmissionMetaData metaData = (MulticastTransmissionMetaData)m.getTransmissionMetaData();

		 // don't print anything
		 /*
		   out.println(metaData + " + " + diff.getClass() + " " +
			     diff.getSeconds() + ":" + diff.getNanoSeconds() +
			     " " + transmissionTime);
		 */

		 //MulticastTransmissionMetaData has following fields:
		 // length - length of packet
		 // ipAddr - source ip addr
		 // mAddr - multicast addr

		 // keep track of the total data volume
		 total += metaData.length;

		 // calculate transmission stats
		 count++;
		 sum += transmissionTime;
		 sumSq += (transmissionTime * transmissionTime);

	     } else {
		 // do nothing
	     }

	     return 0;
	 }

	 /**
	  * Called as the last thing.
	  * Useful for doing any processing after output has finished.
	  */
	 public Object end() throws IOException {
	     PrintStream out = (PrintStream)getOutputStream();

	     Timestamp first = index.getFirstTime();
	     Timestamp last = index.getLastTime();
	     Timestamp diff = TimeCalculator.subtractTimestamp(last, first);
	     //out.println("First element time = " + first);
	     //out.println("Last element time = " + last);

	     out.println("transmission time: " + diff);
	     out.println("transmission volume: " + total);

	     double throughput = ((double)total) / diff.getSeconds();
	     out.printf("throughput: %3.6f\n", throughput);
	     out.printf("average: %3.6f\n", average());
	     out.printf("stddev: %3.6f\n", stddev());

	     return total;
	 }

	 private double average() {
	     return sum / count;
	 }

	 private double stddev() {
	     return (sumSq / count) - (average() * average());
	 }

     }

    /**
     * A TixDump process output so Measurements are printed properly
     * It gets called for each IndexItem.
     */
    class TixDump extends DefaultWriter {
	 public long write(IndexItem item, IndexProperties outputProperties) throws IOException {
	     // get the data item as something that can be deserialized
	     DataItem dataItem = item.getDataItem();
	     Object object = dataItem.getObject();

	     // we're outputting to System.out so getOutputStream() 
	     // should return a PrintStream
	     PrintStream out = (PrintStream)getOutputStream();

	     out.print("TixWriter: ");
	     out.print(item.getIndexTimestamp());
	     out.print("    ");
	     out.print(item.getDataTimestamp());
	     out.print("    ");

	     if (object instanceof Measurement) {
		 out.print("measurement = ");
		 out.println(object);
	     } else {
		 out.println(object.getClass() + " = " + object);
	     }

	     return 0;
	 }
    }
	       
}
