// Matrix2D.java

package uk.co.timeindexing.matrix;

import java.util.List;
import java.util.ArrayList;

/**
 * An implementation of a Matrix.
 * <p>
 * This is a first attempt, so it is likely
 * there are improvements to be made.
 * </p>
 */
public class Matrix2D implements Matrix {
    /*
     * A Matrix is made up of a list of ColumnVectors.
     * The nth element of a column represents row n.
     * Choosing ColumnVectors as the main structure rather than RowVectors
     * makes it easier to append ColumnVectors, 
     * which is used in some Matrix applications.
     */
    List columns = null;

    /*
     * The number of columns.
     */
    int columnMax = 0;

    /*
     * The number of rows.
     */
    int rowMax = 0;

    /**
     * Construct a Matrix of size (m,n).
     * All cells will have Bottom _|_ in them.
     */
    public Matrix2D(int rows, int cols) {
	columns = new ArrayList(cols);

	for (int col=1; col <= cols; col++) {
	    ColumnVector newCol = new ColumnVector(rows);
	    columns.add(newCol);
	}

	columnMax = cols;
	rowMax = rows;
    }
	
    /**
     * Construct a Matrix from an existing Matrix.
     * All cells will have the values of the existing Matrix,
     * as the data will be shared.
     * This consturctor is useful for converting a SquareMatrix into a general Matrix.
     */
    public Matrix2D(Matrix m) {
	collectMatrixColumns(m);
    }

    /**
     * The null constructor is exposed for subclasses.
     * It does nothing.
     */
    Matrix2D() { ; }

    /**
     * Get the no of rows in the Matrix.
     */
    public int rows() {
	return rowMax;
    }

    /**
     * Get the no of columns in the Matrix.
     */
    public int columns() {
	return columnMax;
    }

    /**
     * Get the element (i,j) from the matrix
     * The rows, over i, are numbered 1 to n.
     * The columns, over j,  are numbered 1 to m.
     * @throws IndexOutOfBoundsException if the row is not in the range 1 to n, or the column is not
     * in the range 1 to m
     */
    public Number get(int i, int j) throws IndexOutOfBoundsException {
	if (i < 1 || i > rowMax || j < 1 || j > columnMax) {
	    throw new IndexOutOfBoundsException("Can't get element (" + i + ", " + j + ") from a matrix of size (" + rowMax + ", " + columnMax + ")");
	} else {
	    ColumnVector aCol = column(j);

	    return aCol.get(i);
	}
    }

    /**
     * Set the element (i,j) in the matrix, to value v.
     * The rows, over i, are numbered 1 to n.
     * The columns, over j,  are numbered 1 to m.
     * @throws IndexOutOfBoundsException if the row is not in the range 1 to n, or the column is not
     * in the range 1 to m
     * @return the old value at position (i,j)
     */
    public Number set(int i, int j, Number v) throws IndexOutOfBoundsException {
	if (i < 1 || i >  rowMax || j < 1 || j > columnMax) {
	    throw new IndexOutOfBoundsException("Can't get element (" + i + ", " + j + ") from a matrix of size (" + rowMax + ", " + columnMax + ")");
	} else {
	    ColumnVector aCol = column(j);

	    Number oldValue = aCol.set(i, v);

	    return oldValue;
	}
    }
	
    /**
     * Get the ith row.
     */
    public RowVector row(int i) throws IndexOutOfBoundsException {
	if (i < 1 || i > rowMax) {
	    throw new IndexOutOfBoundsException("Can't get row " + i + " from a matrix of size (" + rowMax + ", " + columnMax + ")");
	} else {
	    // visit each of the columns and get the value for row i out of each column
	    // and create a RowVector

	    // allocate a row of the correct size
	    RowVector row = new RowVector(rowMax);

	    for (int column=1; column <= columnMax; column++) {
		Number n = get(i, column);

		row.set(column, n);
	    }		

	    return  row;
	}
    }

    /**
     * Get the jth column.
     */
    public ColumnVector column(int j) throws IndexOutOfBoundsException {
	if (j < 1 || j > columnMax) {
	    throw new IndexOutOfBoundsException("Can't get column " + j + " from a matrix of size (" + rowMax + ", " + columnMax + ")");
	} else {
	    ColumnVector aCol = (ColumnVector)columns.get(j-1);

	    return aCol;
	}
    }

    /**
     * Add this Matrix to another Matrix.
     */
    public Matrix add(Matrix m) throws UnsupportedOperationException {
	// add Matrices if they are thesame size
	if (this.rows() == m.rows() && this.columns() == m.columns()) {
	    Matrix result = new Matrix2D(rows(), columns());

	    // for each column
	    for (int col=1; col <= columnMax; col++) {
		// for each row
		for (int row=1; row <= rowMax; row++) {
		    // Rij = Aij + Bij 
		    Number value = plus(this.get(row, col), m.get(row, col));
		    result.set(row, col, value);
		}

	    }
	    
	    return result;

	} else {
	    throw new UnsupportedOperationException("Can't add 2 Matrices of different sizes");
	}
    }

    /**
     * Subtract from this Matrix another Matrix.
     */
    public Matrix subtract(Matrix m) throws UnsupportedOperationException {
	// subtract Matrices if they are thesame size
	if (this.rows() == m.rows() && this.columns() == m.columns()) {
	    Matrix result = new Matrix2D(rows(), columns());

	    // for each column
	    for (int col=1; col <= columnMax; col++) {
		// for each row
		for (int row=1; row <= rowMax; row++) {
		    // Rij = Aij - Bij 
		    Number value = sub(this.get(row, col), m.get(row, col));
		    result.set(row, col, value);
		}

	    }
	    
	    return result;

	} else {
	    throw new UnsupportedOperationException("Can't add 2 Matrices of different sizes");
	}
    }

    /**
     * Multiply this Matrix by another Matrix.
     */
    public Matrix multiply(Matrix m) {
	if (this.columns() == m.rows()) {
	    // allocate a result of the correct size
	    Matrix result = new Matrix2D(this.rows(), m.columns());

	    // for each column
	    for (int col=1; col <= columnMax; col++) {
		// for each row
		for (int row=1; row <= rowMax; row++) {
		    // get relevant row and column
		    RowVector r = this.row(row);
		    ColumnVector c = m.column(col);

		    // Rij = sum(vector_multiply(row, col))
		    Number value = vectorMultiply(r, c).sum();

		    // put the value in the correct place in the result
		    result.set(row, col, value);
		}

	    }
	    
	    return result;

	    

	    

	} else {
	    throw new UnsupportedOperationException("Can't multiply 2 Matrices.");
	}
    }

    /**
     * Multiply this Matrix by a scalar
     */
    public Matrix scalarMultiply(Number scalar) {
	Matrix result = new Matrix2D(rows(), columns());

	// for each column
	for (int col=1; col <= columnMax; col++) {
	    // for each row
	    for (int row=1; row <= rowMax; row++) {
		// Rij = scalar * Mij
		Number value = mult(this.get(row, col), scalar);
		result.set(row, col, value);
	    }

	}
	    
	return result;
    }

    /**
     * Fill the Matrix with zeros.
     * @return this Matrix
     */
    public Matrix zero() {
	// fill the whole matrix with zeros

	// for each column
	for (int col = 1; col <= columnMax; col++) {
	    // for each row
	    for (int row = 1; row <= rowMax; row++) {
	    
		// Mij = 0
		set(row, col, new Integer(0));
	    }
	}

	return this;
    }

    /**
     * Append a column to the end of the Matrix.
     * This changes the Matrix inplace, and adds a new column
     * to the Matrix.
     * The size changes from n X m to n X m+1.
     * @throws UnsupportedOperationException if the no of rows in the ColumnVector
     * is different from the no of rows in the Matrix.
     */
    public Matrix append(ColumnVector col) {
	// if the no of rows are the same then add it
	if (col.size() == rows()) {
	    columns.add(col);
	    columnMax++;

	    return this;
	} else {
	    throw new UnsupportedOperationException("Can't add a column with " + col.size() + " rows, to a Matrix of size(" + rowMax + ", " + columnMax + ")");
	}
    }
	    

    /**
     * Is the matrix square.
     */
    public boolean isSquare() {
	if (rows() == columns()) {
	    return true;
	} else {
	    return false;
	}
    }

    /**
     * Add two Number values.
     * @return n1 + n2
     */
    protected Number plus(Number n1, Number n2) {
	return new Double(n1.doubleValue() + n2.doubleValue());
    }

    /**
     * Subtract two Number values.
     * @return n1 - n2
     */
    protected Number sub(Number n1, Number n2) {
	return new Double(n1.doubleValue() - n2.doubleValue());
    }


    /**
     * Multiply two Number values.
     * @return n1 * n2
     */
    protected Number mult(Number n1, Number n2) {
	return new Double(n1.doubleValue() * n2.doubleValue());
    }

    /**
     * Divide two Number values.
     * @return n1 / n2
     */
    protected Number div(Number n1, Number n2) {
	return new Double(n1.doubleValue() / n2.doubleValue());
    }

    /**
     * Multiply two Vectors.
     * Result is [a1*b1, a2*b2, .. , ai*bi,..]
     */
    protected Vector vectorMultiply(Vector v1, Vector v2) {
	if (v1.size() == v2.size()) {
	    // allocate a result Vector of the right size
	    Vector result = new Vector(v1.size());

	    // skip though each vector, multiplying the values
	    for (int pos=1; pos <= result.size(); pos++) {
		Number value = mult(v1.get(pos), v2.get(pos));

		result.set(pos, value);
	    }

	    return result;
	} else {
	    throw new UnsupportedOperationException("Can't multiply 2 Vectors of different sizes.");
	}
    }

    /**
     * Get a handle on all the ColumnVectors in another Matrix.
     * @return the no of columns collected
     */
    protected int collectMatrixColumns(Matrix m) {
	columns = new ArrayList(m.columns());

	for (int col=1; col <= m.columns(); col++) {
	    columns.add(m.column(col));
	}

	columnMax = m.columns();
	rowMax = m.rows();

	return columnMax;
    }

    /**
     * Get a String verson of a Matrix.
     */
    public String toString() {
	StringBuffer buffer = new StringBuffer(256);

	for (int row=1; row <= rowMax; row++) {
	    buffer.append("| ");

	    for (int col=1; col <= columnMax; col++) {
		buffer.append(showNumber(get(row, col)));
		buffer.append("\t");
	    }

	    buffer.append("|\n");
	}

	return buffer.toString();	    
    }

    /**
     * Show a Number in a nicer way.
     * A quick  attempt to remove insignificant digits in printing
     */
    protected String showNumber(Number n) {
	double doubleValue = n.doubleValue();
	float floatValue = n.floatValue();

	if (doubleValue - floatValue == 0) {
	    return (""+n.intValue());
	} else if (doubleValue - floatValue < 0.0000001) {
	    int intValue = n.intValue();

	    if (floatValue == intValue) {
		return (""+intValue);
	    } else {
		return (""+floatValue);
	    }
	} else {
	    return (""+doubleValue);
	}
    }

}

