// MatrixN.java

package uk.co.timeindexing.matrix;

/**
 * An implementation of a Matrix which is square.
 * It has operations that are only valid for square Matrices.
 */
public class MatrixN extends Matrix2D implements SquareMatrix {
    /**
     * Construct a SquareMatrix, where the no of rows and
     * columns is n.
     * @param n the no of rows and columns.
     */
    public MatrixN(int n) {
	super(n,n);
    }

    /**
     * Construct a MatrixN from an existing Matrix,
     * if the existing Matrix is square.
     * All cells will have the values of the existing Matrix,
     * as the data will be shared.
     * This consturctor is useful for converting a general Matrix into a SquareMatrix.
     * @throws UnsupportedOperationException if the Matrix passed in is not square.
     */
    public MatrixN(Matrix m) {
	if (m.isSquare()) {
	    collectMatrixColumns(m);
	} else {
	    throw new UnsupportedOperationException("Can't make a non square matrix into square matrix");
	}
    }


    /**
     * Evaluate the inverse of this Matrix.
     * Uses the following approach:
     * 1/det(M)  * adjoint(M)
     * <p>
     * Using this approach then all the other useful methods
     * for a SquareMatrix drop out due to the codependency of those methods.
     * <p>
     * The inverse, M<sup>-1</sup>, has the property:
     * <blockquote>  M<sup>-1</sup> M = M  M<sup>-1</sup> = I  </blockquote>
     * @return A new Matrix with the inverse
     */
    public SquareMatrix invert() throws MatrixNotInvertableException {
	Number determinant = determinant();

	if (determinant.doubleValue() == 0) {
	    // can;t invert a Matrix with a 0 determinant
	    throw new MatrixNotInvertableException("Can;t invert a Matrix with a 0 determinant");
	} else {
	    // evaluate the inverse
	    Matrix inverse =  adjoint().scalarMultiply(div(new Integer(1), determinant));

	    // now make it into a SquareMatrix
	    SquareMatrix square = new MatrixN(inverse);

	    return square;
	}
    }

    /**
     * Transpose this Matrix.
     * @return A new transposed Matrix
     */
    public SquareMatrix transpose() {
	SquareMatrix result = new MatrixN(size());

	// for each column
	for (int col=1; col <= columnMax; col++) {
	    // for each row
	    for (int row=1; row <= rowMax; row++) {
		// Rji = Mij
		Number value = this.get(row, col);
		result.set(col, row, value);
	    }

	}
	    
	return result;
    }

    /**
     * Evaluate the determinant of this Matrix.
     */
    public Number determinant() {
	if (size() == 1) {
	    return get(1,1);
	} else if (size() == 2) {
	    // special case for a 2 X 2
	    // M11 * M22 - M21 * M12
	    return sub(mult(get(1,1), get(2,2)), mult(get(2,1), get(1,2)));
	} else {
	    // there generic N X N case
	    // M11 * C11 + M12 * C12 + .... + M1n * C1n
	    // Cij = (-1)^i+j * det(minor(i,j))

	    // allocate a Vector to collect up the terms
	    Vector terms = new Vector(rows());

	    for (int col=1; col<=columns(); col++) {
		Number value = get(1,col);

		if (value.doubleValue() == 0) {
		    // no need to get the cofactor, as 0 X cofactor = 0
		    terms.set(col, new Integer(0));
		    continue;
		} else {
		    Number cofactor = cofactor(1,col);
		    terms.set(col, mult(value, cofactor));
		}
	    }

	    // we have a vector of terms
	    // now sum them
	    return terms.sum();
	}
    }

    /**
     * Evaluate the cofactors of this Matrix.
     * @return A new Matrix with the cofactors
     */
    public SquareMatrix cofactors() {
	SquareMatrix result = new MatrixN(size());

	// for each column
	for (int col=1; col <= columnMax; col++) {
	    // for each row
	    for (int row=1; row <= rowMax; row++) {

		// Rij = cofactors(i,j);
		result.set(row,col, cofactor(row,col));
	    }
	}

	return result;
    }

    /**
     * Get the cofactor of  row i and column j.
     * @param i the ith row
     * @param j the jth row
     */
    public Number cofactor(int i, int j) {
	// cofactor of (i,j) is (-1)^i+j * determinant(minor(i,j))

	Number minorDet = minor(i,j).determinant();

	// No need to multiply -1 loads of times
	int sign = (i+j)%2 == 1 ? -1 : 1;

	return mult(new Integer(sign), minorDet);	
    }

    /**
     * Evaluate the adjoint of this Matrix.
     * @return A new Matrix with the adjoint
     */
    public SquareMatrix adjoint() {
	return cofactors().transpose();
    }

    /**
     * Generate the minor of this Matrix.
     * The minor of this matrix is created by 
     * deleting row i and column j.
     * The resulting matrix id of size n-1 X n-1.
     * @param i the ith row
     * @param j the jth row
     */
    public SquareMatrix minor(int i, int j) {
	SquareMatrix result = new MatrixN(size()-1);

	// for each column
	for (int col=1; col <= columnMax; col++) {
	    // for each row
	    for (int row=1; row <= rowMax; row++) {
		// get value, drop relevant row and column
		Number value = this.get(row, col);

		// set the value in the result
		// we need to modify the row and col values
		// as we have removed a row and a column
		if (row < i) {
		    // top half
		    if (col < j) {
			// top left quadrant
			result.set(row, col, value);
		    } else if (col == j) {
			continue;
		    } else {
			// top right quadrant
			result.set(row, col-1, value);
		    }
		} else  if (row == i) {
		    continue;
		} else {
		    // bottom half
		    if (col < j) {
			// bottom left quadrant
			result.set(row-1, col, value);
		    } else if (col == j) {
			continue;
		    } else {
			// bottom right quadrant
			result.set(row-1, col-1, value);
		    }
		}			
	    }

	}
	    
	return result;
    }


    /**
     * Is the matrix square.
     * Always true for a SquareMatrix.
     */
    public boolean isSquare() {
	return true;
    }

    /**
     * Get the size of the Matrix.
     * A square matrix has the same no of rows as columns.
     */
    public int size() {
	return rows();
    }

    /**
     * Append a column to the end of the Matrix is not supported for SquareMatrices,
     * as they will then not be square.
     * @throws UnsupportedOperationException always
     */
    public Matrix append(ColumnVector col) {
	throw new UnsupportedOperationException("Can't add a column to a SquareMatrix");
    }
}
